import json
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Union

from mashumaro import DataClassDictMixin
from mashumaro.mixins.json import DataClassJSONMixin

from .exceptions import UserNotificationException
from .logging import logger
from .subprocess import SubprocessExecutor, which  # nosec


@dataclass
class ScoopFileElement(DataClassDictMixin):
    name: str = field(metadata={"alias": "Name"})
    source: str = field(metadata={"alias": "Source"})

    def __hash__(self) -> int:
        return hash(self.name + self.source)

    def __str__(self) -> str:
        return f"({self.source},{self.name})"


@dataclass
class ScoopInstallConfigFile(DataClassJSONMixin):
    buckets: List[ScoopFileElement]
    apps: List[ScoopFileElement]

    @property
    def bucket_names(self) -> List[str]:
        return [bucket.name for bucket in self.buckets]

    @property
    def app_names(self) -> List[str]:
        return [app.name for app in self.apps]

    @classmethod
    def from_file(cls, scoop_file: Path) -> "ScoopInstallConfigFile":
        with open(scoop_file) as f:
            return cls.from_dict(json.load(f))

    def to_file(self, scoop_file: Path) -> None:
        with open(scoop_file, "w") as f:
            json.dump(self.to_dict(), f, indent=4)


@dataclass
class InstalledApp:
    #: App name
    name: str
    #: App version
    version: str
    #: App root directory
    path: Path
    #: List of bin directories relative to the app path
    bin_dirs: List[Path]


@dataclass
class InstalledScoopApp(InstalledApp):
    #: App scoop manifest file
    manifest_file: Path


class ScoopWrapper:
    def __init__(self) -> None:
        self.scoop_executable = self._find_scoop_executable()
        self.scoop_root_dir = self._find_scoop_root_dir(self.scoop_executable)
        self.logger = logger.bind()

    @property
    def apps_directory(self) -> Path:
        return self.scoop_root_dir.joinpath("apps")

    def install(self, scoop_file: Path) -> List[InstalledScoopApp]:
        """Install scoop apps from a scoop file.
        It returns a list with all apps required to be installed as installed apps."""
        return self.do_install(
            ScoopInstallConfigFile.from_file(scoop_file), self.get_installed_apps()
        )

    def _find_scoop_executable(self) -> Path:
        scoop_path = which("scoop")
        if not scoop_path:
            raise UserNotificationException(
                "Scoop not found in PATH. Please run the build script again."
            )
        return scoop_path

    def _find_scoop_root_dir(self, scoop_executable_path: Path) -> Path:
        pattern = r"^(.*?/scoop/)"
        match = re.match(pattern, scoop_executable_path.absolute().as_posix())

        if match:
            return Path(match.group(1))
        else:
            raise UserNotificationException(
                f"Could not determine scoop directory for {scoop_executable_path}."
            )

    def parse_bin_dirs(
        self, bin_data: Union[str, List[Union[str, List[str]]]]
    ) -> List[Path]:
        """Parse the bin directory from the manifest file."""

        def get_parent_dir(bin_entry: Union[str, List[str]]) -> Optional[Path]:
            bin_path = (
                Path(bin_entry[0]) if isinstance(bin_entry, list) else Path(bin_entry)
            )
            return bin_path.parent if len(bin_path.parts) > 1 else None

        result = []
        if isinstance(bin_data, str):
            result = [parent for parent in [get_parent_dir(bin_data)] if parent]
        elif isinstance(bin_data, list):
            result = [
                parent
                for parent in [get_parent_dir(bin_entry) for bin_entry in bin_data]
                if parent
            ]
        return list(set(result))

    def parse_manifest_file(self, manifest_file: Path) -> InstalledScoopApp:
        app_directory: Path = manifest_file.parent
        tool_name: str = app_directory.parent.name
        with open(manifest_file) as f:
            manifest_data: Dict[str, Any] = json.load(f)
            tool_version: str = manifest_data.get("version", None)
            bin_dirs: List[Path] = self.parse_bin_dirs(manifest_data.get("bin", []))
            return InstalledScoopApp(
                name=tool_name,
                version=tool_version,
                path=app_directory,
                manifest_file=manifest_file,
                bin_dirs=bin_dirs,
            )

    def get_installed_apps(self) -> List[InstalledScoopApp]:
        installed_tools: List[InstalledScoopApp] = []
        self.logger.info(f"Looking for installed apps in {self.apps_directory}")
        manifest_files = list(self.apps_directory.glob("*/*/manifest.json"))

        with ThreadPoolExecutor() as executor:
            future_to_file = {
                executor.submit(self.parse_manifest_file, manifest_file): manifest_file
                for manifest_file in manifest_files
            }
            for future in as_completed(future_to_file):
                installed_tools.append(future.result())

        return installed_tools

    def do_install(
        self,
        scoop_install_config: ScoopInstallConfigFile,
        installed_apps: List[InstalledScoopApp],
    ) -> List[InstalledScoopApp]:
        """Install scoop apps from a scoop file."""

        newly_installed_apps = self.do_install_missing(
            scoop_install_config, installed_apps
        )
        # If some apps where just installed we need to update the list of installed apps
        if newly_installed_apps:
            self.logger.info(
                "New apps were installed, update the list of installed apps."
            )
            updated_installed_apps = self.get_installed_apps()
        else:
            updated_installed_apps = installed_apps
        apps = self.map_required_apps_to_installed_apps(
            scoop_install_config.app_names, updated_installed_apps
        )
        return apps

    def do_install_missing(
        self,
        scoop_install_config: ScoopInstallConfigFile,
        installed_apps: List[InstalledScoopApp],
    ) -> List[ScoopFileElement]:
        """Check which apps are installed and install the missing ones."""
        apps_to_install = self.get_tools_to_be_installed(
            scoop_install_config, installed_apps
        )
        if not apps_to_install:
            self.logger.info("All Scoop apps already installed. Skip installation.")
            return []
        already_installed_apps = set(scoop_install_config.apps) - set(apps_to_install)
        if already_installed_apps:
            self.logger.info(
                f"Scoop apps already installed: {','.join(str(app) for app in already_installed_apps)}"
            )
        self.logger.info(
            f"Start installing missing apps: {','.join(str(app) for app in apps_to_install)}"
        )

        # Create a temporary scoopfile with the remaining apps to install and install them
        with TemporaryDirectory() as tmp_dir:
            tmp_scoop_file = Path(tmp_dir).joinpath("scoopfile.json")
            ScoopInstallConfigFile(
                scoop_install_config.buckets, apps_to_install
            ).to_file(tmp_scoop_file)
            SubprocessExecutor(
                [self.scoop_executable, "import", tmp_scoop_file]
            ).execute()
        return apps_to_install

    @staticmethod
    def map_required_apps_to_installed_apps(
        app_names: List[str],
        installed_apps: List[InstalledScoopApp],
    ) -> List[InstalledScoopApp]:
        """Map the required apps to the installed apps."""
        # convert the list of installed apps into a dictionary for faster lookup
        installed_apps_dict = {app.name: app for app in installed_apps}
        try:
            apps = [installed_apps_dict[app_name] for app_name in app_names]
        except KeyError as e:
            raise UserNotificationException(
                f"Could not find {e} in the installed apps. Something went wrong during the scoop installation."
            )
        return apps

    @staticmethod
    def get_tools_to_be_installed(
        scoop_install_config: ScoopInstallConfigFile,
        installed_tools: List[InstalledScoopApp],
    ) -> List[ScoopFileElement]:
        """Check which apps are installed and install the missing ones."""
        installed_tools_names = {tool.name for tool in installed_tools}
        return [
            app
            for app in scoop_install_config.apps
            if app.name not in installed_tools_names
        ]



""" Logging utilities. """

import sys
import time
from pathlib import Path
from typing import Any, Callable, Optional

from loguru import logger as _logger

from .docs_utils import fulfills

logger = _logger

# Adding custom logging levels
logger.level("START", no=38, color="<yellow>")
logger.level("STOP", no=39, color="<yellow>")


@fulfills("REQ-LOGGING_TIME_IT-0.0.1")
def time_it(message: Optional[str] = None) -> Callable[..., Any]:
    """Decorator to time a function."""

    def _time_it(func: Callable[..., Any]) -> Callable[..., Any]:
        text = message or f"{func.__module__}.{func.__qualname__}"

        def time_it(*args: Any, **kwargs: Any) -> Any:
            start_time = time.time()
            logger.log("START", f"Starting {text}")
            result = func(*args, **kwargs)
            end_time = time.time()
            logger.log("STOP", f"Finished {text} in {end_time-start_time:.2f}s")
            return result

        return time_it

    return _time_it


@fulfills("REQ-LOGGING_FILE-0.0.1")
def setup_logger(log_file: Optional[Path] = None, clear: bool = True) -> None:
    """Setup logger to stdout and optionally to file."""
    logger.remove()
    logger.add(
        sys.stdout,
        format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
        "<level>{level: <8}</level> | <level>{message}</level>",
    )
    if log_file is not None:
        logger.add(log_file, level="DEBUG")
        # Clear log file
        if log_file.exists() and clear:
            log_file.write_text("")



from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path

from mashumaro import DataClassDictMixin

from yanga.core.cmd_line import Command, register_arguments_for_config_dataclass
from yanga.core.logging import logger, time_it
from yanga.core.scoop_wrapper import ScoopWrapper


@dataclass
class InstallCommandConfig(DataClassDictMixin):
    scoop_file: Path = field(metadata={"help": "Scoop install configuration file."})

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "InstallCommandConfig":
        return cls.from_dict(vars(namespace))


class InstallCommand(Command):
    def __init__(self) -> None:
        super().__init__(
            "install", "Run a yanga installer based on the provided arguments"
        )
        self.logger = logger.bind()

    @time_it("Install")
    def run(self, args: Namespace) -> int:
        self.logger.debug(f"Running {self.name} with args {args}")
        self.do_run(InstallCommandConfig.from_namespace(args))
        return 0

    def do_run(self, config: InstallCommandConfig) -> int:
        ScoopWrapper().install(config.scoop_file)
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, InstallCommandConfig)



from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, OrderedDict, TypeAlias

import yaml
from mashumaro import DataClassDictMixin


@dataclass
class StageConfig(DataClassDictMixin):
    #: Stage name or class name if file is not specified
    stage: str
    #: Path to file with stage class
    file: Optional[str] = None
    #: Stage class name
    class_name: Optional[str] = None
    #: Stage description
    description: Optional[str] = None
    #: Stage timeout in seconds
    timeout_sec: Optional[int] = None


PipelineConfig: TypeAlias = OrderedDict[str, List[StageConfig]]


@dataclass
class YangaConfig(DataClassDictMixin):
    #: Pipeline stages
    pipeline: PipelineConfig

    @classmethod
    def from_file(cls, config_file: Path) -> "YangaConfig":
        with open(config_file) as fs:
            config_dict = yaml.safe_load(fs)
        return cls.from_dict(config_dict)



from pathlib import Path
from typing import List

from yanga.core.logging import logger
from yanga.core.scoop_wrapper import ScoopWrapper
from yanga.ybuild.environment import BuildEnvironment
from yanga.ybuild.pipeline import Stage


class YangaScoopInstall(Stage):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        super().__init__(environment, group_name)
        self.logger = logger.bind()

    def get_name(self) -> str:
        return "yanga_scoop_install"

    @property
    def scoop_file(self) -> Path:
        return self.environment.project_root_dir.joinpath("scoopfile.json")

    def run(self) -> int:
        self.logger.info(
            f"Run {self.__class__.__name__} stage. Output dir: {self.output_dir}"
        )
        ScoopWrapper().install(self.scoop_file)
        return 0

    def get_inputs(self) -> List[Path]:
        return [self.scoop_file]

    def get_outputs(self) -> List[Path]:
        return []



from dataclasses import dataclass
from pathlib import Path

from yanga.ybuild.project import ProjectBuildArtifactsLocator


@dataclass
class BuildEnvironment:
    variant_name: str
    build_config: str
    build_target: str
    project_root_dir: Path

    @property
    def artifacts_locator(self) -> ProjectBuildArtifactsLocator:
        return ProjectBuildArtifactsLocator(
            self.project_root_dir, self.variant_name, self.build_config
        )



# create a Runnable protocol and make Executor accept it
import hashlib
import json
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import List

from .logging import logger


class Runnable(ABC):
    @abstractmethod
    def run(self) -> int:
        """Run stage"""

    @abstractmethod
    def get_name(self) -> str:
        """Get stage name"""

    @abstractmethod
    def get_inputs(self) -> List[Path]:
        """Get stage dependencies"""

    @abstractmethod
    def get_outputs(self) -> List[Path]:
        """Get stage outputs"""


class RunInfoStatus(Enum):
    MATCH = (False, "Nothing changed. Previous execution info matches.")
    NO_INFO = (True, "No previous execution info found.")
    FILE_NOT_FOUND = (True, "File not found.")
    FILE_CHANGED = (True, "File has changed.")

    def __init__(self, should_run: bool, message: str) -> None:
        self.should_run = should_run
        self.message = message


class Executor:
    """Accepts Runnable objects and executes them.
    It create a file with the same name as the runnable's name
    and stores the inputs and outputs with their hashes.
    If the file exists, it checks the hashes of the inputs and outputs
    and if they match, it skips the execution."""

    RUN_INFO_FILE_EXTENSION = ".deps.json"

    def __init__(self, cache_dir: Path) -> None:
        self.cache_dir = cache_dir

    @staticmethod
    def get_file_hash(path: Path) -> str:
        with open(path, "rb") as file:
            bytes = file.read()
            readable_hash = hashlib.sha256(bytes).hexdigest()
            return readable_hash

    def store_run_info(self, runnable: Runnable) -> None:
        file_info = {
            "inputs": {
                str(path): self.get_file_hash(path) for path in runnable.get_inputs()
            },
            "outputs": {
                str(path): self.get_file_hash(path) for path in runnable.get_outputs()
            },
        }

        run_info_path = self.get_runnable_run_info_file(runnable)
        run_info_path.parent.mkdir(parents=True, exist_ok=True)
        with run_info_path.open("w") as f:
            # pretty print the json file
            json.dump(file_info, f, indent=4)

    def get_runnable_run_info_file(self, runnable: Runnable) -> Path:
        return self.cache_dir / f"{runnable.get_name()}{self.RUN_INFO_FILE_EXTENSION}"

    def previous_run_info_matches(self, runnable: Runnable) -> RunInfoStatus:
        run_info_path = self.get_runnable_run_info_file(runnable)
        if not run_info_path.exists():
            return RunInfoStatus.NO_INFO

        with run_info_path.open() as f:
            previous_info = json.load(f)

        for file_type in ["inputs", "outputs"]:
            for path_str, previous_hash in previous_info[file_type].items():
                path = Path(path_str)
                if not path.exists():
                    return RunInfoStatus.FILE_NOT_FOUND
                elif self.get_file_hash(path) != previous_hash:
                    return RunInfoStatus.FILE_CHANGED
        return RunInfoStatus.MATCH

    def execute(self, runnable: Runnable) -> int:
        run_info_status = self.previous_run_info_matches(runnable)
        if run_info_status.should_run:
            logger.info(
                f"Runnable '{runnable.get_name()}' must run. {run_info_status.message}"
            )
            exit_code = runnable.run()
            self.store_run_info(runnable)
            return exit_code
        logger.info(
            f"Runnable '{runnable.get_name()}' execution skipped. {run_info_status.message}"
        )

        return 0



from yanga.core.logging import logger
from yanga.ybuild.config import YangaConfig
from yanga.ybuild.environment import BuildEnvironment
from yanga.ybuild.pipeline import PipelineLoader, StageRunner


class YangaBuild:
    def __init__(self, environment: BuildEnvironment) -> None:
        self.logger = logger.bind()
        self.environment = environment
        self.artifacts_locator = environment.artifacts_locator

    def run(self) -> None:
        project_dir = self.artifacts_locator.project_root_dir.absolute().as_posix()
        self.logger.info(f"Run yanga build in '{project_dir}'")
        project_config = YangaConfig.from_file(self.artifacts_locator.config_file)
        stages = PipelineLoader(
            project_config.pipeline, self.artifacts_locator.project_root_dir
        ).load_stages()
        for stage in stages:
            StageRunner(self.environment, stage).run()



from dataclasses import dataclass, field
from typing import List, Optional

from mashumaro.mixins.yaml import DataClassYAMLMixin


@dataclass
class VariantConfig(DataClassYAMLMixin):
    """
    .. item:: IMPL-VariantConfig
        :fulfills:
            REQ-VARIANT_NAME-0.0.1
            REQ-VARIANT_SLUG-0.0.1
            REQ-VARIANT_DESCRIPTION-0.0.1
            REQ-VARIANT_TAGS-0.0.1
    """

    name: str
    description: Optional[str] = None
    tags: List[str] = field(default_factory=list)
    _slug: Optional[str] = None

    def generate_slug(self, slug: Optional[str] = None) -> str:
        return (slug or self.name).lower().replace(" ", "-")

    @property
    def slug(self) -> str:
        return self._slug or self.generate_slug()

    @slug.setter
    def slug(self, slug: str) -> None:
        self._slug = self.generate_slug(slug)



import sys
from argparse import ArgumentParser
from sys import argv

from yanga import __version__
from yanga.commands.build import BuildCommand
from yanga.commands.init import InitCommand
from yanga.commands.install import InstallCommand
from yanga.core.cmd_line import CommandLineHandlerBuilder
from yanga.core.exceptions import UserNotificationException
from yanga.core.logging import logger, setup_logger


def do_run() -> None:
    parser = ArgumentParser(prog="yanga", description="Yanga CLI", exit_on_error=False)
    parser.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    builder = CommandLineHandlerBuilder(parser)
    builder.add_commands([BuildCommand(), InitCommand(), InstallCommand()])
    handler = builder.create()
    handler.run(argv[1:])


def main() -> int:
    try:
        setup_logger()
        do_run()
    except UserNotificationException as e:
        logger.error(f"{e}")
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



from abc import ABC
from dataclasses import dataclass
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import List, Type

from yanga.core.exceptions import UserNotificationException
from yanga.core.runnable import Executor, Runnable
from yanga.ybuild.config import PipelineConfig, StageConfig
from yanga.ybuild.environment import BuildEnvironment


class Stage(Runnable, ABC):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        self.environment = environment
        self.output_dir = self.environment.artifacts_locator.build_dir / group_name


@dataclass
class BuildStage:
    group_name: str
    _class: Type[Stage]


class PipelineLoader:
    def __init__(self, pipeline_config: PipelineConfig, project_root_dir: Path) -> None:
        self.pipeline_config = pipeline_config
        self.project_root_dir = project_root_dir

    def load_stages(self) -> List[BuildStage]:
        result = []
        for group_name, stages_config in self.pipeline_config.items():
            result.extend(
                self._load_stages(group_name, stages_config, self.project_root_dir)
            )
        return result

    @staticmethod
    def _load_stages(
        group_name: str, stages_config: List[StageConfig], project_root_dir: Path
    ) -> List[BuildStage]:
        result = []
        for stage_config in stages_config:
            stage_class_name = stage_config.class_name or stage_config.stage
            if not stage_config.file:
                # no file means that the stage is a built-in stage
                stage_class = PipelineLoader._load_builtin_stage(stage_class_name)
            else:
                stage_class = PipelineLoader._load_user_stage(
                    project_root_dir.joinpath(stage_config.file), stage_class_name
                )
            result.append(BuildStage(group_name, stage_class))
        return result

    @staticmethod
    def _load_builtin_stage(stage_class_name: str) -> Type[Stage]:
        stage_module = import_module("yanga.ybuild.stages")
        try:
            stage_class = getattr(stage_module, stage_class_name)
        except AttributeError:
            raise UserNotificationException(
                f"Stage '{stage_class_name}' is not a Yanga built-in stage."
                " Please check your pipeline configuration."
            )
        return stage_class

    @staticmethod
    def _load_user_stage(python_file: Path, stage_class_name: str) -> Type[Stage]:
        # Create a module specification from the file path
        spec = spec_from_file_location(f"user__{stage_class_name}", python_file)
        if spec and spec.loader:
            stage_module = module_from_spec(spec)
            # Import the module
            spec.loader.exec_module(stage_module)
            try:
                stage_class = getattr(stage_module, stage_class_name)
            except AttributeError:
                raise UserNotificationException(
                    f"Could not load class '{stage_class_name}' from file '{python_file}'."
                    " Please check your pipeline configuration."
                )
            return stage_class
        raise UserNotificationException(
            f"Could not load file '{python_file}'."
            " Please check the file for any errors."
        )


class StageRunner:
    """It checks if a stage must run in current environment.
    A stage shall run if any of the dependencies changed or one of the outputs is missing.
    All dependencies and outputs relevant information is stored in the stage output directory
    in a <stage>.deps file.
    """

    def __init__(self, environment: BuildEnvironment, stage: BuildStage) -> None:
        self.environment = environment
        self.stage = stage

    def run(self) -> None:
        stage = self.stage._class(self.environment, self.stage.group_name)
        Executor(stage.output_dir).execute(stage)



from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path

from mashumaro import DataClassDictMixin

from yanga.core.cmd_line import Command, register_arguments_for_config_dataclass
from yanga.core.logging import logger, time_it
from yanga.ybuild.build_main import YangaBuild
from yanga.ybuild.environment import BuildEnvironment


@dataclass
class BuildCommandConfig(DataClassDictMixin):
    variant_name: str = field(metadata={"help": "SPL variant name."})
    build_config: str = field(metadata={"help": "Build configuration name."})
    build_target: str = field(metadata={"help": "Build target name."})
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={
            "help": "Project root directory. "
            "Defaults to the current directory if not specified."
        },
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "BuildCommandConfig":
        return cls.from_dict(vars(namespace))


class BuildCommand(Command):
    def __init__(self) -> None:
        super().__init__("build", "Build a yanga project")
        self.logger = logger.bind()

    @time_it("Build")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = BuildCommandConfig.from_namespace(args)
        # search variant
        # create project build artifacts locator
        environment = BuildEnvironment(
            config.variant_name,
            config.build_config,
            config.build_target,
            config.project_dir,
        )
        YangaBuild(environment).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, BuildCommandConfig)
